import sys
import cx_Oracle
import os
import itertools
import json
from collections import defaultdict
import csv
import tempfile
import pickle
from datetime import timedelta, datetime
import re

dir = sys.argv[1]
TAG_RE = re.compile(r'<[^>]+>')


def remove_tags(text):
    return TAG_RE.sub('', text)


def load_disciplines():
    with open(os.path.join(dir, 'mapping_discipline.csv'), encoding="Latin-1") as f:
        reader = csv.DictReader(f)
        return dict([d['formal_name'].lower().replace("'",''), d['disc_name']] for d in reader)


def load_nih_data():
    print("Loading NIH data")
    with open(os.path.join(dir, 'NIH_22Aug2017projects 2012-2017-Indiana-Univ-Purdue.csv')) as f:
        reader = csv.DictReader(f)
        projects = [line for line in reader]

    return projects


def load_nsf_data():
    print("Loading NSF data")
    with open(os.path.join(dir, 'copi-nsf-11-20.csv')) as f:
        reader = csv.DictReader(f)
        projects = [line for line in reader]

    for project in projects:
        project['Project Number'] = project['Project']
        co_pis = project['coPI'].replace(',', ';').strip()
        if not co_pis:
            co_pis = 'Not Applicable'
        project['Other PI or Project Leader(s)'] = co_pis
        project['Contact PI / Project Leader'] = project['First.Name'] + ' ' + project['Last.Name']
    return projects


def load_grant_data():
    grants = data_cache('grants.pkl')
    print("Loading grant data")
    if not grants:
        query = """select lower(a.PERSON_USER_ID), a.CGAWD_TOT_AMT, a.KC_AWD_NBR, CG_AGENCY_RPT_NM, CG_AGENCY_AWD_NBR
        from DSS_KC.KC_AWD_ALL_PRSN_V a
        where a.CGAWD_BEG_DT > DATE'2012-01-01' and a.kc_awd_node = 'B' and a.CGAWD_TOT_AMT > 0"""
        remote = cx_Oracle.connect("hpastats", os.environ['GRANTS_PW'],
                                   "esdbd100.uits.indiana.edu:1521/DSS1PRD.uits.indiana.edu")
        remotecursor = remote.cursor()
        remotecursor.execute(query)
        grants = remotecursor.fetchall()
        write_cache('grants.pkl', grants)

    return grants


def data_cache(file):
    cachefile = os.path.join(tempfile.gettempdir(), file)
    try:
        cache_time = datetime.fromtimestamp(os.path.getmtime(cachefile))
        if datetime.now() - cache_time < timedelta(weeks=1):
            with open(cachefile, 'rb') as f:
                return pickle.load(f)
    except os.error:
        return None


def write_cache(file, data):
    cachefile = os.path.join(tempfile.gettempdir(), file)
    with open(cachefile, 'wb') as f:
        pickle.dump(data, f)

def load_storage_data():
    """SQL query: select user, gb gigabytes, system from
((select user, sum(gigabytes) gb, 'SDA' system
from user_storages where system = 'hpss' and date = '2017-08-30' group by user)
union
(select user, sum(gigabytes) gb, 'DC2' system from user_storages where system = 'DC2' and date = '2017-05-14'
group by user)) c
where gb >= 1
order by gb"""
    with open(os.path.join(dir, 'storage.csv'), encoding="Latin-1") as f:
        reader = csv.DictReader(f)
        return [line for line in reader]

def load_job_data():
    jobs = data_cache('jobs.pkl')
    print("Loading job data")
    if not jobs:
        core_seconds = ("task_count * (to_number(extract(second from duration)) + "
                        "       to_number(extract(minute from duration)) * 60 + "
                        "       to_number(extract(hour from duration)) * 3600 + "
                        "       to_number(extract(day from duration)) * 86400 )")

        query = """select cluster_name, user_id, %s as core_seconds
        from iursa.job_transaction where submit_time > DATE'2012-01-01'""" % core_seconds
        remote = cx_Oracle.connect("hpastats", os.environ['HPASTATS_PW'],
                                   "rdc03.uits.iu.edu:1521/iursa.iu.edu")
        remotecursor = remote.cursor()
        remotecursor.execute(query)
        jobs = remotecursor.fetchall()
        write_cache('jobs.pkl', jobs)

    return jobs


def matching(pub, projects):
    for project_id in projects.keys():
        if pub in project_id:
            return projects[project_id]


def normalize_journal(j):
    result = j.split(':')[0]
    result = result.lstrip('The ')
    result = result.replace('.', '').lower()
    result = result.replace('&', 'and')
    result = result.replace('(amsterdam, netherlands)', '')
    result = result.replace(chr(39), '')    # ??
    result = result.strip()
    return result


def projects_of_hpc_users(iu_grants, nih_projects, hpc_users):
    """Returns a dictionary (IU user ID) => (Grant Information)"""
    print("Matching publications")
    hpc_projects = defaultdict(list)

    grant_dict = dict([g[4], g] for g in iu_grants if g[4])
    for project in nih_projects:
        core_project_number = project['Project Number']
        grant = matching(core_project_number, grant_dict)
        copis = project['Other PI or Project Leader(s)']
        if grant and copis != 'Not Applicable':
            user = grant[0]
            if user in hpc_users:
                pi = project['Contact PI / Project Leader']
                proj = dict(GrantID=grant[4],
                             GrantSource=grant[3],
                             GrantSize=grant[1],
                             CoPIS=[pi.strip()] + [c.strip() for c in copis.split(';')]
                             )
                hpc_projects[user].append(proj)
    return hpc_projects


def get_pi_projects(pi, pi_pairs):
    pi_projects = []
    for pair, projects in pi_pairs.items():
        if pi in pair:
            pi_projects += projects
    return pi_projects

def write_schema(pi_ids, pi_pairs):
    node_schema = [dict(name='name', type='string'), dict(name='id', type='numeric'),
                   dict(name='total_amount', type='numeric'), dict(name='number_of_grants', type='numeric')]
    edge_schema = [dict(name='source', type='numeric'), dict(name='target', type='numeric'),
                   dict(name='total_amount', type='numeric'), dict(name='number_of_grants', type='numeric')]
    result = dict(edges=dict(schema=edge_schema), nodes=dict(schema=node_schema))

    result['nodes']['data'] = list()
    for pi, id in pi_ids.items():
        pi_projects = get_pi_projects(id, pi_pairs)

        if pi_projects:
            result['nodes']['data'].append(dict(id=id,
                                    name=pi,
                                    number_of_grants=len(pi_projects),
                                    total_amount=sum(p['GrantSize'] for p in pi_projects)
                                    ))

    edges = []
    for pair, projects in pi_pairs.items():
        a = list(pair)
        edges.append(dict(source=a[0],
                          target=a[1],
                          number_of_grants=len(projects),
                          total_amount=sum(p['GrantSize'] for p in projects))
                     )

    result['edges']['data'] = edges

    print(json.dumps(result))


def load_hpc_users():
    hpc_users = set()
    print("Analyzing jobs")
    jobs = load_job_data()
    for job in jobs:
        if job[0] != 'BIGRED2PLUS':
            hpc_users.add(job[1])

    print("Analyzing storage")
    storage = load_storage_data()
    for s in storage:
        hpc_users.add(s['user'])

    return hpc_users


def main():
    projects = load_nsf_data() + load_nih_data()
    grants = load_grant_data()
    hpc_users = load_hpc_users()

    hpc_user_projects = projects_of_hpc_users(grants, projects, hpc_users)

    project_list = list(p for p in itertools.chain.from_iterable(hpc_user_projects.values()) if p['GrantSize'] > 0)

    pis = list(itertools.chain.from_iterable(r['CoPIS'] for r in project_list))
    pi_ids = dict([author, index] for index, author in enumerate(pis))

    pi_pairs = defaultdict(list)
    for project in project_list:
        pairs = list(itertools.product(project['CoPIS'], project['CoPIS']))
        for pair in pairs:
            if pair[0] != pair[1]:
                key = frozenset([pi_ids[pair[0]], pi_ids[pair[1]]])
                pi_pairs[key].append(project)

    write_schema(pi_ids, pi_pairs)
    # print(json.dumps(result))

if __name__ == '__main__':
    main()